from asciimatics.event import KeyboardEvent
from asciimatics.scene import Scene
from asciimatics.screen import Screen
from asciimatics.exceptions import ResizeScreenError, StopApplication, NextScene
import sys
import src.models as models
import src.views as views


# Event handler for global keys
def global_shortcuts(event):
    if isinstance(event, KeyboardEvent):
        key = event.key_code
        if key == Screen.KEY_F1:
            raise NextScene("Main")
        elif key == Screen.KEY_F2:
            raise NextScene("Commits")
        elif key == Screen.KEY_F3:
            raise NextScene("Working Copy")
        elif key == Screen.ctrl("a"):
            raise NextScene("Shortcuts")
        elif key == Screen.ctrl("x"):
            raise StopApplication("User terminated app")


def main(screen, scene):
    scenes = [
        Scene([views.BranchListView(screen, branch_model)], -1, name="Main"),
        Scene([views.CommitView(screen, commit_model)], -1, name="Commits"),
        Scene([views.CommitFilesView(screen, commit_model)], -1, name="View Commit Details"),
        Scene([views.CommitFileDiffView(screen, commit_model)], -1, name="View Commit Diff"),
        Scene([views.WorkingCopyView(screen, working_copy_model)], -1, name="Working Copy"),
        Scene([views.ShortcutsView(screen)], -1, name="Shortcuts"),
        Scene([views.ExceptionView(screen, branch_model)], -1, name="Error")
    ]

    screen.play(scenes, stop_on_resize=True, start_scene=scene, allow_int=True, unhandled_input=global_shortcuts)


branch_model = models.GitBranchModel()
commit_model = models.GitCommitModel()
working_copy_model = models.WorkingCopyModel()
last_scene = None
while True:
    try:
        Screen.wrapper(main, catch_interrupt=True, arguments=[last_scene])
        sys.exit(0)
    except ResizeScreenError as e:
        last_scene = e.scene



# coding: utf-8
import os
from git import Repo
from pydriller import RepositoryMining


class GitModel(object):
    """
    Parent class for all Git models.
    """
    def __init__(self):
        # Current branch when editing.
        self.current_id = None
        self.last_error = None

        self.dir_path = os.getcwd()
        self.repo = Repo(self.dir_path)
        self.branches = self.repo.branches
        assert not self.repo.bare


class GitBranchModel(GitModel):
    """
    Model for listing and manipulating Git branches.
    """
    def list_branches(self):
        """
        Lists branches for Branches table. Prepends the current branch with a tick.
        :return: A list of branch_name, branch tuples.
        """
        branches = []
        for branch in self.branches:
            branch_name = branch.name
            if self.is_current_branch(branch):
                branch_name = "✔ " + branch_name
            else:
                branch_name = "  " + branch_name
            if branch.tracking_branch() is not None:
                branch_name = branch_name + " -> " + branch.tracking_branch().name
            branch_with_name = [branch_name, branch]
            branches.append(branch_with_name)
        return branches

    def is_current_branch(self, branch):
        return branch == self.get_current_branch()

    def get_current_branch(self):
        return self.repo.active_branch

    def get_current_contact(self):
        if self.current_id is None:
            return {"name": "", "address": "", "phone": "", "email": "", "notes": ""}
        else:
            return self.get_contact(self.current_id)

    def checkout_branch(self, branch_name):
        # self.repo.git.stash()
        self.repo.git.checkout(branch_name)
        # self.repo.git.stash('pop')


class GitCommitModel(GitModel):
    def __init__(self):
        super(GitCommitModel, self).__init__()
        self.repository_miner = RepositoryMining(self.dir_path)
        self.current_commit = None
        self.current_commit_file = None

    """
    A model for listing and manipulating git commits.
    """
    def list_commits(self):
        """
        Lists commits for the current branch for the commits table.
        :return: An array of arrays with commit information inside.
        """
        _commits_with_info = []
        for commit in reversed(list(self.repository_miner.traverse_commits())):
            _commit_info = [commit.hash[:12], commit.msg, commit.author.name,
                            commit.author_date.strftime("%d/%m/%Y, %H:%M:%S")]
            _commit_info_with_hash_as_key = [_commit_info, commit]
            _commits_with_info.append(_commit_info_with_hash_as_key)
        return _commits_with_info

    def list_files_in_current_commit(self):
        _file_map = []
        if self.current_commit is not None:
            for diff in self.current_commit.modifications:
                _file_map.append([diff.new_path, diff])
        return _file_map

    def current_file_diff(self):
        _diff = ""
        if self.current_commit_file is not None:
            _diff = self.current_commit_file.diff
        return _diff


class WorkingCopyModel(GitModel):
    """
    A model for viewing working copy file changes.
    """
    def list_of_changed_files(self):
        return self.repo.head.commit.diff(None)

    def list_of_changed_unadded_files(self):
        return self.repo.index.diff(None)

    def list_of_changed_files_in_index(self):
        return self.repo.index.diff('HEAD')

    def changed_files_for_table(self):
        """
        Lists changed files in the working copy. Prepends files added to the index with a tick.
        :return: A list of formatted_file_path, file_path tuples representing changed working copy files.
        """
        changed_files = []
        for file in self.list_of_changed_files_in_index():
            _file_name_with_diff = ["✔ " + file.b_path, file.b_path]
            changed_files.append(_file_name_with_diff)
        for file in self.list_of_changed_unadded_files():
            _file_name_with_diff = ["  " + file.b_path, file.b_path]
            changed_files.append(_file_name_with_diff)
        for file in self.repo.untracked_files:
            _file_name_with_diff = ["  " + file, file]
            changed_files.append(_file_name_with_diff)
        return changed_files

    def toggle_add_file_to_index(self, path):
        added = [item.a_path for item in self.list_of_changed_files_in_index()]
        if path in added:
            self.repo.git.reset("--", path)
        else:
            self.repo.git.add(path)

    def commit(self, commit_message):
        self.repo.git.commit('-m {}'.format(commit_message))

    def push(self):
        origin = self.repo.remote(name='origin')
        origin.push()



from asciimatics.widgets import Frame, ListBox, Layout, Divider, Text, \
    Button, TextBox, Widget, MultiColumnListBox
from asciimatics.exceptions import NextScene, StopApplication
from git import GitCommandError


class View(Frame):
    def __init__(self, model, screen, height, width, data=None, on_load=None,
                 has_border=True, hover_focus=False, name=None, title=None,
                 x=None, y=None, has_shadow=False, reduce_cpu=False, is_modal=False,
                 can_scroll=True):
        super(View, self).__init__(screen=screen,
                                   height=height,
                                   width=width,
                                   data=data,
                                   on_load=on_load,
                                   has_border=has_border,
                                   hover_focus=hover_focus,
                                   name=name,
                                   title=title,
                                   x=x,
                                   y=y,
                                   has_shadow=has_shadow,
                                   reduce_cpu=reduce_cpu,
                                   is_modal=is_modal,
                                   can_scroll=can_scroll)
        self._model = model
        self.nav_header = None
        self.set_theme("tlj256")

    def add_navigation_header(self):
        self.nav_header = Layout([1, 1, 1, 1])
        self.add_layout(self.nav_header)
        self.nav_header.add_widget(Button("Branches", self._open_branch_view), 0)
        self.nav_header.add_widget(Button("History", self._open_commits_view), 1)
        self.nav_header.add_widget(Button("Commit", self._open_working_copy_view), 2)
        self.nav_header.add_widget(Button("Quit", self._quit), 3)

    def add_divider(self):
        divider_layout = Layout([100])
        self.add_layout(divider_layout)
        divider_layout.add_widget(Divider())

    def add_shortcut_panel(self):
        layout0 = Layout([100])
        _header = TextBox(1, as_string=True)
        _header.disabled = True
        _header.custom_colour = "label"
        _header.value = "Press ctrl-a to see a list of shortcuts. Press ctrl-x to quit."
        self.add_layout(layout0)
        layout0.add_widget(_header, 0)

    @staticmethod
    def _open_branch_view():
        raise NextScene("Main")

    @staticmethod
    def _open_commits_view():
        raise NextScene("Commits")

    @staticmethod
    def _open_working_copy_view():
        raise NextScene("Working Copy")

    @staticmethod
    def _quit():
        raise StopApplication("User pressed quit")


class BranchListView(View):
    def __init__(self, screen, model):
        super(BranchListView, self).__init__(model,
                                             screen,
                                             screen.height * 9 // 10,
                                             screen.width * 9 // 10,
                                             on_load=self._reload_list,
                                             hover_focus=True,
                                             can_scroll=False,
                                             title="Branches")
        self._model = model

        # Create the form for displaying the list of branches.
        self._list_view = ListBox(
            Widget.FILL_FRAME,
            model.list_branches(),
            name="branches",
            add_scroll_bar=True,
            on_select=self.__checkout)

        self.add_navigation_header()
        self.add_divider()

        self.table_layout = Layout([100], fill_frame=True)
        self.add_layout(self.table_layout)
        self.table_layout.add_widget(self._list_view)
        self.table_layout.add_widget(Divider())

        self.add_shortcut_panel()
        self.fix()

    def _reload_list(self):
        self.nav_header.blur()
        self.table_layout.focus()
        self._list_view.options = self._model.list_branches()
        self._list_view.value = self._model.get_current_branch()

    def __checkout(self):
        self.save()
        try:
            self._model.checkout_branch(self.data["branches"])
        except GitCommandError as e:
            self._model.last_error = e
            raise NextScene("Error")
        self._model.current_id = self.data["branches"]
        self._reload_list()


class CommitView(View):
    def __init__(self, screen, model):
        super(CommitView, self).__init__(model,
                                         screen,
                                         screen.height * 9 // 10,
                                         screen.width * 9 // 10,
                                         on_load=self._reload_list,
                                         hover_focus=True,
                                         can_scroll=False,
                                         title="Commits")
        self._model = model

        self._list_view = MultiColumnListBox(
            Widget.FILL_FRAME,
            ["10%", "60%", "15%", "15%"],
            model.list_commits(),
            titles=["Hash", "Message", "Author", "Date"],
            name="commits",
            add_scroll_bar=True,
            on_select=self.__view_diff)

        self.add_navigation_header()
        self.add_divider()

        self.table_layout = Layout([100], fill_frame=True)
        self.add_layout(self.table_layout)
        self.table_layout.add_widget(self._list_view)
        self.table_layout.add_widget(Divider())

        self.add_shortcut_panel()
        self.fix()

    def __view_diff(self):
        self.save()
        self._model.current_commit = self.data["commits"]
        self.__open_commit_details_scene()

    @staticmethod
    def __open_commit_details_scene():
        raise NextScene("View Commit Details")

    def _reload_list(self):
        self._list_view.options = self._model.list_commits()
        self.nav_header.blur()
        self.table_layout.focus()


class CommitFilesView(View):
    def __init__(self, screen, model):
        super(CommitFilesView, self).__init__(model,
                                              screen,
                                              screen.height * 9 // 10,
                                              screen.width * 9 // 10,
                                              on_load=self._reload_list,
                                              hover_focus=True,
                                              can_scroll=False,
                                              title="Diff Files")
        self._model = model

        # Create the form for displaying the list of branches.
        self._list_view = ListBox(
            Widget.FILL_FRAME,
            model.list_files_in_current_commit(),
            name="diff_files",
            add_scroll_bar=True,
            on_select=self.__view_diff)

        self.add_navigation_header()
        self.add_divider()

        self.table_layout = Layout([100], fill_frame=True)
        self.add_layout(self.table_layout)
        self.table_layout.add_widget(self._list_view)

        layout2 = Layout([1, 1, 1, 1])
        self.add_layout(layout2)
        layout2.add_widget(Button("Back", self._open_commits_view), 0)

        self.table_layout.add_widget(Divider())

        self.add_shortcut_panel()
        self.fix()

    def __view_diff(self):
        self.save()
        self._model.current_commit_file = self.data["diff_files"]
        self.__open_commit_file_diff_scene()

    @staticmethod
    def __open_commit_file_diff_scene():
        raise NextScene("View Commit Diff")

    def _reload_list(self):
        self._list_view.options = self._model.list_files_in_current_commit()
        self.nav_header.blur()
        self.table_layout.focus()


class CommitFileDiffView(View):
    def __init__(self, screen, model):
        super(CommitFileDiffView, self).__init__(model,
                                            screen,
                                            screen.height * 9 // 10,
                                            screen.width * 9 // 10,
                                            hover_focus=True,
                                            can_scroll=False,
                                            title="Commit Diff",
                                            reduce_cpu=True)
        self._model = model
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        self._header = TextBox(20, as_string=True)
        self._header.disabled = True
        self._header.custom_colour = "label"
        layout.add_widget(self._header)
        layout2 = Layout([1, 1, 1, 1])
        self.add_layout(layout2)
        layout2.add_widget(Button("Back", self._open_commits_view), 0)
        self.fix()

    def reset(self):
        # Do standard reset to clear out form, then populate with new data.
        super(CommitFileDiffView, self).reset()
        self._header.value = self._model.current_file_diff()


class WorkingCopyView(View):
    def __init__(self, screen, model):
        super(WorkingCopyView, self).__init__(model,
                                              screen,
                                              screen.height * 9 // 10,
                                              screen.width * 9 // 10,
                                              on_load=self._reload_list,
                                              hover_focus=True,
                                              can_scroll=False,
                                              title="Working copy")
        self._model = model
        screen.catch_interrupt = True

        self._list_view = ListBox(
            Widget.FILL_FRAME,
            model.changed_files_for_table(),
            name="working_copy",
            add_scroll_bar=True,
            on_select=self._add_to_index)

        self.add_navigation_header()
        self.add_divider()

        self.table_layout = Layout([100], fill_frame=True)
        self.add_layout(self.table_layout)
        self.table_layout.add_widget(self._list_view)
        self.table_layout.add_widget(Divider())

        self._commit_button = Button("Commit", self._commit)
        self._push_button = Button("Push", self._push)

        layout2 = Layout([8, 1, 1])
        self.add_layout(layout2)

        self.commit_message = Text("Commit message:", "commit_message")
        layout2.add_widget(self.commit_message, 0)
        layout2.add_widget(self._commit_button, 1)
        layout2.add_widget(self._push_button, 2)

        self.add_divider()

        self.add_shortcut_panel()
        self.fix()

    def _add_to_index(self):
        self.save()
        self._model.toggle_add_file_to_index(self.data["working_copy"])
        self._reload_list()

    def _commit(self):
        self._model.commit(self.commit_message.value)
        self.reset()

    def _push(self):
        self._model.push()
        self.reset()

    def reset(self):
        # Do standard reset to clear out form, then populate with new data.
        super(WorkingCopyView, self).reset()
        self._reload_list()

    def _reload_list(self):
        self._list_view.options = self._model.changed_files_for_table()
        self.nav_header.blur()
        self.table_layout.focus()


class ExceptionView(View):
    def __init__(self, screen, model):
        super(ExceptionView, self).__init__(model,
                                            screen,
                                            screen.height * 9 // 10,
                                            screen.width * 9 // 10,
                                            hover_focus=True,
                                            can_scroll=False,
                                            title="Error",
                                            reduce_cpu=True)
        self._model = model
        layout = Layout([100], fill_frame=True)
        self.add_layout(layout)
        self._header = TextBox(20, as_string=True)
        self._header.disabled = True
        self._header.custom_colour = "label"
        layout.add_widget(self._header)
        layout2 = Layout([1, 1, 1, 1])
        self.add_layout(layout2)
        layout2.add_widget(Button("OK", self._ok), 0)
        self.fix()

    def reset(self):
        # Do standard reset to clear out form, then populate with new data.
        super(ExceptionView, self).reset()
        self._header.value = "Error encountered: {0}".format(self._model.last_error)

    def _ok(self):
        raise NextScene("Main")

    @staticmethod
    def _cancel():
        raise NextScene("Main")


class ShortcutsView(Frame):
    def __init__(self, screen):
        super(ShortcutsView, self).__init__(screen,
                                            screen.height * 6 // 10,
                                            screen.width * 6 // 10,
                                            hover_focus=True,
                                            can_scroll=False,
                                            title="Shortcuts",
                                            reduce_cpu=True)
        self.set_theme("tlj256")

        layout = Layout([100, 100], fill_frame=True)
        self.add_layout(layout)
        self._row0 = TextBox(20, as_string=True)
        self._row0.disabled = True
        self._row0.value = "F1 - See branch window\n" \
                           "Ctrl-A - See shortcuts window"
        layout.add_widget(self._row0, 0)
        self._row1 = TextBox(20, as_string=True)
        self._row1.disabled = True
        self._row1.value = "F2 - See commit list window\n" \
                           "F3 - See working copy and commit window"
        layout.add_widget(self._row1, 1)
        layout2 = Layout([1, 1])
        self.add_layout(layout2)
        layout2.add_widget(Button("Okay", self._ok), 1)
        self.fix()

    def reset(self):
        super(ShortcutsView, self).reset()

    def _ok(self):
        raise NextScene("Main")

    @staticmethod
    def _cancel():
        raise NextScene("Main")



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup (
    name="asciigit",
    description="A command line ASCII Git GUI",
    long_description=long_description,
    long_description_content_type='text/markdown',
    url="https://github.com/tommyclark/asciigit",
    version="0.0.6",
    packages=find_packages(),
    tests_require=['mock'],
    install_requires=[
        'asciimatics',
        'GitPython',
        'pydriller'
    ],
    entry_points={
        "console_scripts": [
            "asciigit = src.main:main",
        ],
    }
)



# coding: utf-8
import unittest
from src.models import GitBranchModel, GitCommitModel, WorkingCopyModel
from tests.fixtures import *


class TestGitBranchModel(unittest.TestCase):
    def setUp(self):
        self.branch_model = GitBranchModel.__new__(GitBranchModel)
        self.branch_model.repo = MockRepository()

    def testGetActiveBranch(self):
        branch = self.branch_model.get_current_branch()
        assert branch.name == "sausage_branch"

    def testIsActiveBranch(self):
        assert not (self.branch_model.is_current_branch(MockNonTrackingHead()))
        assert not (self.branch_model.is_current_branch(None))
        assert self.branch_model.is_current_branch(self.branch_model.get_current_branch())

    def testCheckout(self):
        assert self.branch_model.is_current_branch(MockCurrentHead())
        self.branch_model.checkout_branch(MockNonTrackingHead())
        assert self.branch_model.is_current_branch(MockNonTrackingHead())

    def testListBranches(self):
        self.branch_model.branches = self.branch_model.repo.branches

        branch_1 = ['✔ sausage_branch -> sausage_branch', MockCurrentHead()]
        branch_2 = ['  bacon_branch -> bacon_branch', MockAlternateHead()]
        branch_3 = ['  not_tracking_branch', MockNonTrackingHead()]
        expected_list = [branch_1, branch_2, branch_3]

        branches = self.branch_model.list_branches()
        assert branches is not None
        assert branches == expected_list


class TestCommitModel(unittest.TestCase):
    def setUp(self):
        self.commit_model = GitCommitModel.__new__(GitCommitModel)
        self.commit_model.repository_miner = MockRepositoryMiner()
        self.commit_model.current_model = None

    def testListCommits(self):
        commits = self.commit_model.list_commits()
        assert commits is not None

        commit_info1 = [['hash1', 'msg1', 'name1', '03/03/1999, 00:00:00'], commits[3][1]]
        commit_info2 = [['hash2', 'msg2', 'name2', '03/03/1999, 00:00:00'], commits[2][1]]
        commit_info3 = [['hash3', 'msg3', 'name3', '03/03/1999, 00:00:00'], commits[1][1]]
        commit_info4 = [['hash4', 'msg4', 'name4', '03/03/1999, 00:00:00'], commits[0][1]]
        commit_with_info_list = [commit_info4, commit_info3, commit_info2, commit_info1]

        assert commits == commit_with_info_list

    def testListFilesInCurrentCommit(self):
        date = datetime(1999, 3, 3)
        commit = MockRepositoryMinerCommit("hash1", "msg1", "name1", date)
        self.commit_model.current_commit = commit
        files = self.commit_model.list_files_in_current_commit()

        assert files == [["test_path", commit.modifications[0]],
                         ["alt_test_path", commit.modifications[1]]]

    def testRetrievingCurrentDiff(self):
        self.commit_model.current_commit_file = MockRepositoryMinerCommitModification()
        modification = self.commit_model.current_file_diff()
        assert modification == "test_diff"


class TestWorkingCopyModel(unittest.TestCase):
    def setUp(self):
        self.working_copy_model = WorkingCopyModel.__new__(WorkingCopyModel)
        self.working_copy_model.repo = MockRepository()

    def testListChangedFiles(self):
        changed_files = self.working_copy_model.list_of_changed_unadded_files()

        assert changed_files is not None
        assert changed_files[0].b_path == "unadded_4"
        assert changed_files[1].b_path == "unadded_5"

    def testListUnaddedFiles(self):
        changed_files = self.working_copy_model.list_of_changed_files()

        assert changed_files is not None
        assert changed_files[0].b_path == "test_1"
        assert changed_files[1].b_path == "test_2"
        assert changed_files[2].b_path == "test_3"
        assert changed_files[3].b_path == "unadded_4"
        assert changed_files[4].b_path == "unadded_5"

    def testListOfChangedFilesInIndex(self):
        changed_index_files = self.working_copy_model.list_of_changed_files_in_index()

        assert changed_index_files is not None
        assert changed_index_files[0].b_path == "test_1"
        assert changed_index_files[1].b_path == "test_2"
        assert changed_index_files[2].b_path == "test_3"

    def testInfoForTable(self):
        test_info = self.working_copy_model.changed_files_for_table()

        assert test_info is not None
        changed_file1 = ['✔ test_1', 'test_1']
        changed_file2 = ['✔ test_2', 'test_2']
        changed_file3 = ['✔ test_3', 'test_3']
        changed_file4 = ['  unadded_4', 'unadded_4']
        changed_file5 = ['  unadded_5', 'unadded_5']
        changed_file6 = ['  untracked_file', 'untracked_file']
        changed_files = [changed_file1, changed_file2, changed_file3, changed_file4, changed_file5, changed_file6]

        assert changed_files == test_info

    def testTogglingAddFiles(self):
        added_file = "test_1"
        self.working_copy_model.toggle_add_file_to_index(added_file)
        assert added_file in self.working_copy_model.repo.git.reset_files
        assert added_file not in self.working_copy_model.repo.git.added_files

        unadded_file = "unadded_4"
        self.working_copy_model.toggle_add_file_to_index(unadded_file)
        assert unadded_file in self.working_copy_model.repo.git.added_files
        assert unadded_file not in self.working_copy_model.repo.git.reset_files

    def testCommit(self):
        self.working_copy_model.commit("Hello there")
        assert self.working_copy_model.repo.git.message == "-m Hello there"

    def testPush(self):
        self.working_copy_model.push()
        assert self.working_copy_model.repo.remote("origin").pushed
